                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/

#ifndef _MODEL_H_
#define _MODEL_H_

#include "../Common/Set.h"
#include "../Common/Tokenizer.h"

struct ArrayVertex {
	Vertex vertex;
	Vertex normal;
};

struct ModelVertex {
	Vertex vertex;
	Vertex normal;
	float s,t;
};


bool operator == (ModelVertex &mv0, ModelVertex &mv1){
	return (mv0.vertex == mv1.vertex) && (mv0.normal == mv1.normal) && (mv0.s == mv1.s) && (mv0.t == mv1.t);
}


/* -------------------------- */


void getMem(ArrayVertex *&vertexArray, float *&texCoordArray, int nVertices, int nKeyFrames){
	vertexArray = (ArrayVertex *) malloc(nVertices * (nKeyFrames * sizeof(ArrayVertex) + 2 * sizeof(float)));
	texCoordArray = (float *) (((char *) vertexArray) + nVertices * nKeyFrames * sizeof(ArrayVertex));
}

void freeMem(ArrayVertex *vertexArray){
	free(vertexArray);
}

/* -------------------------- */

class KeyFrame {
protected:
	Set <ModelVertex> modelVertices;
	Set <int> indices;

	bool loadFromGof(char *fileName, float scale);
public:
	KeyFrame(char *fileName, float scale){
		loadFromGof(fileName, scale);
	}

	Set <ModelVertex> *getVertices(){
		return &modelVertices;
	}

	Set <int> *getIndices(){
		return &indices;
	}

};

bool KeyFrame::loadFromGof(char *fileName, float scale){
	Set <Vertex> vertices;
	Set <Vertex> normals;
	Set <float> texCoords;
	Set <int> vertexIndexArray;
	Set <int> normalIndexArray;
	Set <int> texCoordIndexArray;
	Set <int> *cIndexArray;

	Vertex v;
	int i,j,k,index,first;

	char str[256];
	Tokenizer tok;
	tok.setFile(fileName);

	while (tok.next(str)){
		if (stricmp(str, "vertices") == 0){
			while (tok.next(str) && *str != '>'){
				if (stricmp(str, "coord") == 0){
					for (i = 0; i < 3; i++){
						tok.next(str);
						if (*str == '-'){
							tok.next(str);
							v[i] = -(float) atof(str);
						} else v[i] = (float) atof(str);
					}
					vertices.add(scale * v);
				}
			}
		} else if (stricmp(str, "normals") == 0){
			while (tok.next(str) && *str != '<');

			while (tok.next(str) && *str != '>'){
				for (i = 0; i < 3; i++){
					if (*str == '-'){
						tok.next(str);
						v[i] = -(float) atof(str);
					} else v[i] = (float) atof(str);
					tok.next(str);
				}
				normals.add(v);
			}
		} else if (stricmp(str, "uv") == 0){
			for (i = 0; i < 5; i++) tok.next(NULL);

			while (tok.next(str) && *str != '>'){
				if (*str == '<'){
					for (i = 0; i < 2; i++){
						tok.next(str);
						if (*str == '-'){
							tok.next(str);
							texCoords.add(-(float) atof(str));
						} else texCoords.add((float) atof(str));
						tok.next(NULL);
					}
					tok.next(NULL);
				}
			}
		} else if (stricmp(str, "faces") == 0){
			while (tok.next(str) && *str != '<');

			while (tok.next(str) && *str != '>'){
				for (k = 0; k < 3; k++){
					while (tok.next(str) && *str != '<');

					cIndexArray = (k == 0)? &vertexIndexArray : (k == 1)? &normalIndexArray : &texCoordIndexArray;
					i = 0;
					first = cIndexArray->getSize();
					while (tok.next(str) && *str != '>'){
						index = atoi(str);
						if (i > 2){
							cIndexArray->add((*cIndexArray)[first]);
							cIndexArray->add((*cIndexArray)[j]);
						}
						j = cIndexArray->add(index);
						
						tok.next(NULL);
						i++;
					}
				}
				tok.next(NULL);
			}
		}
	}




	ModelVertex mv;

	for (i = 0; i < vertexIndexArray.getSize(); i++){
		mv.vertex = vertices[vertexIndexArray[i]];
		mv.vertex.x = -mv.vertex.x;
		mv.normal = normals [normalIndexArray[i]];
		mv.normal.x = -mv.normal.x;
		mv.s  =     texCoords[2 * texCoordIndexArray[i]];
		mv.t  = 1 - texCoords[2 * texCoordIndexArray[i] + 1];

		for (j = modelVertices.getSize() - 1; j >= 0; j--){
			if (mv == modelVertices[j])	break;				
		}
		if (j < 0){
			j = modelVertices.add(mv);
		}
		indices.add(j);
	}


	//CString string, tmp;
/*	for (i = 0; i < vertices.getSize(); i++){
		tmp.Format("%g %g %g\n", vertices[i].x, vertices[i].y, vertices[i].z);
		string += tmp;
	}*/
/*	for (i = 0; i < normals.getSize(); i++){
		tmp.Format("%g %g %g\n", normals[i].x, normals[i].y, normals[i].z);
		string += tmp;
	}*/
/*	for (i = 0; i < texCoords.getSize(); i += 2){
		tmp.Format("%g %g\n", texCoords[i], texCoords[i+1]);
		string += tmp;
	}
*/
/*	for (i = 0; i < vertexIndexArray.getSize(); i += 3){
		tmp.Format("%d %d %d\n", vertexIndexArray[i], vertexIndexArray[i+1], vertexIndexArray[i+2]);
		string += tmp;
	}*/
	/*for (i = 0; i < normalIndexArray.getSize(); i += 3){
		tmp.Format("%d %d %d\n", normalIndexArray[i], normalIndexArray[i+1], normalIndexArray[i+2]);
		string += tmp;
	}*/
	/*for (i = 0; i < texCoordIndexArray.getSize(); i += 3){
		tmp.Format("%d %d %d\n", texCoordIndexArray[i], texCoordIndexArray[i+1], texCoordIndexArray[i+2]);
		string += tmp;
	}*/


	//string.Format("%d %d", modelVertices.getSize(), indices.getSize());

	//::MessageBox(NULL, string, "mu", MB_OK);



	return true;
}



/* -------------------------- */

class Model {
protected:
	ArrayVertex *vertexArray;
	int nVertices;
	float *texCoordArray;


// Using OpenGL
#ifdef __GL_H__
	GLuint vaoBuffer;
#else
	LPDIRECT3DVERTEXBUFFER8 texCoordsVB;
	LPDIRECT3DVERTEXBUFFER8 *keyFramesVB;
	LPDIRECT3DINDEXBUFFER8  indexBuffer;
	DWORD blendShader;
#endif

	int *indexArray;
	int nIndices;

	int nKeyFrames;

	TextureID textureID;
public:
	Model();
	~Model();

	void setTexture(TextureID texture){
		textureID = texture;
	}

	bool loadKeyFrames(Set <KeyFrame *> &keyFrames);

	void saveToFile(char *fileName);
	void loadFromFile(char *fileName);

	void draw(int keyFrame0, int keyFrame1, float weight, TextureHandler *texHandler);

};

Model::Model(){
	vertexArray = NULL;
	nVertices = 0;

	indexArray = NULL;
	nIndices = 0;

	nKeyFrames = 0;

#ifdef __GL_H__
	vaoBuffer = -1;
#else
	texCoordsVB = NULL;
	keyFramesVB = NULL;
	indexBuffer = NULL;
#endif
}

Model::~Model(){
#ifdef __GL_H__
	if (vaoBuffer != -1) glDeleteObjectBufferATI(vaoBuffer);
#else
	if (texCoordsVB != NULL) texCoordsVB->Release();
	for (int i = 0; i < nKeyFrames; i++){
		if (keyFramesVB[i] != NULL) keyFramesVB[i]->Release();
	}
	if (keyFramesVB != NULL) delete keyFramesVB;

	if (indexBuffer != NULL) indexBuffer->Release();
	d3ddevice->DeleteVertexShader(blendShader);
#endif

	if (vertexArray != NULL) freeMem(vertexArray);
	if (indexArray  != NULL) delete indexArray;
}

bool Model::loadKeyFrames(Set <KeyFrame *> &keyFrames){
	Set <ModelVertex> *modelVertices;
	Set <int> *indices;

	int i, j;
	nKeyFrames = keyFrames.getSize();

	if (nKeyFrames > 0){
		modelVertices = keyFrames[0]->getVertices();
		indices       = keyFrames[0]->getIndices();

		nVertices = modelVertices->getSize();
		nIndices  = indices->getSize();

		getMem(vertexArray, texCoordArray, nVertices, nKeyFrames);

		indexArray = new int[nIndices];
		for (i = 0; i < nIndices; i++){
			indexArray[i] = (*indices)[i];
		}

		for (i = 0; i < nVertices; i++){
			texCoordArray[2 * i    ] = (*modelVertices)[i].s;
			texCoordArray[2 * i + 1] = (*modelVertices)[i].t;
		}
		
	} else return false;

	ArrayVertex *dest = vertexArray;
	for (i = 0; i < nKeyFrames; i++){
		modelVertices = keyFrames[i]->getVertices();
		
		for (j = 0; j < nVertices; j++){
			dest->vertex = (*modelVertices)[j].vertex;
			dest->normal = (*modelVertices)[j].normal;
			dest++;
		}

		delete keyFrames[i];
	}



/*
	CString string, tmp;

	for (i = 0; i < nVertices; i++){
		tmp.Format("%g %g %g\n", vertexArray[i].vertex.x, vertexArray[i].vertex.y, vertexArray[i].vertex.z);
		string += tmp;
	}

	string += "\n\n";

	for (i = 0; i < nIndices; i += 3){
		tmp.Format("%d %d %d\n", indexArray[i], indexArray[i+1], indexArray[i+2]);
		string += tmp;
	}

	::MessageBox(NULL, string, "mu", MB_OK);
*/




	return true;
}

void Model::saveToFile(char *fileName){
	FILE *file = fopen(fileName, "wb");

	fwrite(&nVertices,    sizeof(nVertices), 1, file);
	fwrite(&nIndices,     sizeof(nIndices),  1, file);
	fwrite(&nKeyFrames,   sizeof(nKeyFrames),1, file);
	fwrite(vertexArray,   sizeof(ArrayVertex) * nVertices * nKeyFrames, 1, file);
	fwrite(texCoordArray, sizeof(float)   * 2 * nVertices, 1, file);
	fwrite(indexArray,    sizeof(int)         * nIndices,  1, file);

	fclose(file);
}

void Model::loadFromFile(char *fileName){
	FILE *file = fopen(fileName, "rb");

	fread(&nVertices,    sizeof(nVertices), 1, file);
	fread(&nIndices,     sizeof(nIndices),  1, file);
	fread(&nKeyFrames,   sizeof(nKeyFrames),1, file);

	getMem(vertexArray, texCoordArray, nVertices, nKeyFrames);
	indexArray = new int[nIndices];

	fread(vertexArray,   sizeof(ArrayVertex) * nVertices * nKeyFrames, 1, file);
	fread(texCoordArray, sizeof(float)   * 2 * nVertices, 1, file);
	fread(indexArray,    sizeof(int)         * nIndices,  1, file);

	fclose(file);

	/*char str[256];
	sprintf(str, "%d", nIndices / 3);
	::MessageBox(NULL, str, "mu", MB_OK);
	*/




// Using OpenGL
#ifdef __GL_H__

	if (GL_ATI_vertex_array_object){
		vaoBuffer = glNewObjectBufferATI(nVertices * (sizeof(ArrayVertex) * nKeyFrames + sizeof(float) * 2), vertexArray, GL_STATIC_ATI);
	}

// Using Direct3D
#else
	d3ddevice->CreateIndexBuffer(nIndices * sizeof(int), D3DUSAGE_WRITEONLY, D3DFMT_INDEX32, D3DPOOL_MANAGED, &indexBuffer);
	
	unsigned char *dest;
	indexBuffer->Lock(0, nIndices * sizeof(int), &dest, 0);
	memcpy(dest, indexArray, nIndices * sizeof(int));
	indexBuffer->Unlock();
	
	d3ddevice->CreateVertexBuffer(nVertices * 2 * sizeof(float), D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &texCoordsVB);
	
	texCoordsVB->Lock(0, nVertices * 2 * sizeof(float), &dest, 0);
	memcpy(dest, texCoordArray, nVertices * 2 * sizeof(float));	
	texCoordsVB->Unlock();

	keyFramesVB = new LPDIRECT3DVERTEXBUFFER8[nKeyFrames];
	for (int i = 0; i < nKeyFrames; i++){
		d3ddevice->CreateVertexBuffer(nVertices * sizeof(ArrayVertex), D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, keyFramesVB + i);
		
		keyFramesVB[i]->Lock(0, nVertices * sizeof(ArrayVertex), &dest, 0);
		memcpy(dest, vertexArray + i * nVertices, nVertices * sizeof(ArrayVertex));
		keyFramesVB[i]->Unlock();
	}

	DWORD shader[] = {
		D3DVSD_STREAM(0),
		D3DVSD_REG(D3DVSDE_POSITION,   D3DVSDT_FLOAT3),
		D3DVSD_REG(D3DVSDE_NORMAL,     D3DVSDT_FLOAT3),
		D3DVSD_STREAM(1),
		D3DVSD_REG(D3DVSDE_TEXCOORD0,  D3DVSDT_FLOAT2),
		D3DVSD_STREAM(2),
		D3DVSD_REG(D3DVSDE_POSITION2,  D3DVSDT_FLOAT3),
		D3DVSD_REG(D3DVSDE_NORMAL2,    D3DVSDT_FLOAT3),
		
		D3DVSD_END()
	};
	
	d3ddevice->CreateVertexShader(shader, NULL, &blendShader, 0);


#endif


}


// Using OpenGL
#ifdef __GL_H__

void Model::draw(int keyFrame0, int keyFrame1, float weight, TextureHandler *texHandler){

	texHandler->setTextures(textureID);
	//texHandler->setVertexShader(0);
	texHandler->apply();

	texHandler->setActiveTexture(0);


	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	glEnable(GL_LIGHTING);

	glEnable(GL_LIGHT0);
	GLfloat Ambient[4]     = {0, 0, 0, 0};
	GLfloat LightColor0[4] = {1, 1, 1, 1};
	GLfloat LightPos0[4]   = {1, 1, 1, 0};

    glLightfv(GL_LIGHT0, GL_POSITION, LightPos0);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  LightColor0);
    glLightfv(GL_LIGHT0, GL_AMBIENT,  Ambient);


	float matrix[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, matrix);

	
	
	glEnable(GL_VERTEX_BLEND_ARB);

	glMatrixMode(GL_MODELVIEW1_ARB);
	glVertexBlendEnviATI(GL_VERTEX_SOURCE_ATI, GL_VERTEX_STREAM0_ATI);

	glMatrixMode(GL_MODELVIEW1_ARB);
	glLoadMatrixf(matrix);

	glVertexBlendEnviATI(GL_VERTEX_SOURCE_ATI, GL_VERTEX_STREAM1_ATI);


	glVertexBlendARB(2);
	GLfloat weights[2] = {weight, 1 - weight};
	glWeightfvARB(2, weights);



	if (GL_ATI_vertex_array_object){
		glArrayObjectATI(GL_TEXTURE_COORD_ARRAY, 2, GL_FLOAT, 2 * sizeof(float), vaoBuffer, nVertices * nKeyFrames * sizeof(ArrayVertex));
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glClientActiveVertexStreamATI(GL_VERTEX_STREAM0_ATI);
		glArrayObjectATI(GL_VERTEX_ARRAY, 3, GL_FLOAT, sizeof(ArrayVertex), vaoBuffer, keyFrame0 * nVertices * sizeof(ArrayVertex));
		glArrayObjectATI(GL_NORMAL_ARRAY, 3, GL_FLOAT, sizeof(ArrayVertex), vaoBuffer, keyFrame0 * nVertices * sizeof(ArrayVertex) + sizeof(Vertex));

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);


		glClientActiveVertexStreamATI(GL_VERTEX_STREAM1_ATI);
		glArrayObjectATI(GL_VERTEX_ARRAY, 3, GL_FLOAT, sizeof(ArrayVertex), vaoBuffer, keyFrame1 * nVertices * sizeof(ArrayVertex));
		glArrayObjectATI(GL_NORMAL_ARRAY, 3, GL_FLOAT, sizeof(ArrayVertex), vaoBuffer, keyFrame1 * nVertices * sizeof(ArrayVertex) + sizeof(Vertex));

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);

	} else {
		glTexCoordPointer(2, GL_FLOAT, sizeof(float) * 2, texCoordArray);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		
		glClientActiveVertexStreamATI(GL_VERTEX_STREAM0_ATI);
		glVertexPointer(  3, GL_FLOAT, sizeof(ArrayVertex),            vertexArray + keyFrame0 * nVertices);
		glNormalPointer(     GL_FLOAT, sizeof(ArrayVertex), ((char *) (vertexArray + keyFrame0 * nVertices)) + sizeof(Vertex));
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		
		glClientActiveVertexStreamATI(GL_VERTEX_STREAM1_ATI);
		glVertexPointer(  3, GL_FLOAT, sizeof(ArrayVertex),            vertexArray + keyFrame1 * nVertices);
		glNormalPointer(     GL_FLOAT, sizeof(ArrayVertex), ((char *) (vertexArray + keyFrame1 * nVertices)) + sizeof(Vertex));
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
	}



	
	glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, indexArray);




	glClientActiveVertexStreamATI(GL_VERTEX_STREAM1_ATI);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	glClientActiveVertexStreamATI(GL_VERTEX_STREAM0_ATI);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glDisable(GL_VERTEX_BLEND_ARB);



	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHTING);

	glDisable(GL_CULL_FACE);


	texHandler->apply();

/*	Vertex v;
	glBegin(GL_LINES);
	for (int i = 0; i < nIndices; i++){
		glColor3f(0,1,0);
		v = vertexArray[indexArray[i]].vertex;
		glVertex3fv((GLfloat *) &v);

		v += 7 * vertexArray[indexArray[i]].normal;
		glColor3f(1,0,0);
		glVertex3fv((GLfloat *) &v);
	}
	glEnd();
*/

}

#else // Using Direct3D


void Model::draw(int keyFrame0, int keyFrame1, float weight, TextureHandler *texHandler){

	texHandler->setTextures(textureID);
	texHandler->apply();
	
	d3ddevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
	
	d3ddevice->SetIndices(indexBuffer, 0);
	d3ddevice->SetStreamSource(0, keyFramesVB[keyFrame0], sizeof(ArrayVertex));
	d3ddevice->SetStreamSource(1, texCoordsVB,            sizeof(float) * 2);
	d3ddevice->SetStreamSource(2, keyFramesVB[keyFrame1], sizeof(ArrayVertex));
	d3ddevice->SetVertexShader(blendShader);
	
	d3ddevice->SetRenderState(D3DRS_TWEENFACTOR, *((DWORD *) &weight));
	d3ddevice->SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_TWEENING);
	

	/*float nPatches = 2.0f;
	d3ddevice->SetRenderState(D3DRS_PATCHEDGESTYLE, D3DPATCHEDGE_DISCRETE);
	d3ddevice->SetRenderState(D3DRS_PATCHSEGMENTS, *((DWORD *) &nPatches));
*/

	d3ddevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, nIndices, 0, nIndices / 3);
/*	
	nPatches = 1.0f;
	d3ddevice->SetRenderState(D3DRS_PATCHSEGMENTS, *((DWORD *) &nPatches));

*/	
	d3ddevice->SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_DISABLE);
	
	d3ddevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	
}

#endif // _D3D8_H_

#endif // _MODEL_H_