                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/

#ifndef _EXPRESSION_H_
#define _EXPRESSION_H_

#include <math.h>
#include "Tokenizer.h"

class Expression {
protected:

public:
    Expression(){}
    virtual ~Expression(){}

    virtual float eval() = 0;


};

/* ------------------------------------ */

class Num : public Expression  {
protected:
    float num;
public:
    Num(float number){
        num = number;
    }

    float eval(){
        return num;
    }
};

/* ------------------------------------ */

class Add : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Add(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }
    ~Add(){
        delete exp0;
        delete exp1;
    }


    float eval(){
        return exp0->eval() + exp1->eval();
    }
};

/* ------------------------------------ */

class Sub : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Sub(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }
    ~Sub(){
        delete exp0;
        delete exp1;
    }


    float eval(){
        return exp0->eval() - exp1->eval();
    }
};

/* ------------------------------------ */

class Mul : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Mul(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }
    ~Mul(){
        delete exp0;
        delete exp1;
    }


    float eval(){
        return exp0->eval() * exp1->eval();
    }
};

/* ------------------------------------ */

class Div : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Div(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }
    ~Div(){
        delete exp0;
        delete exp1;
    }


    float eval(){
        return exp0->eval() / exp1->eval();
    }
};

/* ------------------------------------ */

class Cos : public Expression  {
protected:
    Expression *expr;
public:
    Cos(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return cosf(expr->eval());
    }
};

/* ------------------------------------ */

class Sin : public Expression  {
protected:
    Expression *expr;
public:
    Sin(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return sinf(expr->eval());
    }
};

/* ------------------------------------ */

class Tan : public Expression  {
protected:
    Expression *expr;
public:
    Tan(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return tanf(expr->eval());
    }
};

/* ------------------------------------ */

class Exp : public Expression  {
protected:
    Expression *expr;
public:
    Exp(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return expf(expr->eval());
    }
};

/* ------------------------------------ */

class Log : public Expression  {
protected:
    Expression *expr;
public:
    Log(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return logf(expr->eval());
    }
};

/* ------------------------------------ */

class Pow : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Pow(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }

    float eval(){
        return powf(exp0->eval(), exp1->eval());
    }
};

/* ------------------------------------ */

class Sqrt : public Expression  {
protected:
    Expression *expr;
public:
    Sqrt(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return sqrtf(expr->eval());
    }
};

/* ------------------------------------ */

class Abs : public Expression  {
protected:
    Expression *expr;
public:
    Abs(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return fabsf(expr->eval());
    }
};

/* ------------------------------------ */

class Angle : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Angle(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }

    float eval(){
        return atan2f(exp1->eval(), exp0->eval());
    }
};

/* ------------------------------------ */

class Floor : public Expression  {
protected:
    Expression *expr;
public:
    Floor(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return floorf(expr->eval());
    }
};

/* ------------------------------------ */

class Ceil : public Expression  {
protected:
    Expression *expr;
public:
    Ceil(Expression *Expr){
        expr = Expr;
    }

    float eval(){
        return ceilf(expr->eval());
    }
};

/* ------------------------------------ */

class Min : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Min(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }

    float eval(){
        float e0, e1;
        e0 = exp0->eval();
        e1 = exp1->eval();
        return (e0 < e1)? e0 : e1;
    }
};

/* ------------------------------------ */

class Max : public Expression  {
protected:
    Expression *exp0, *exp1;
public:
    Max(Expression *Exp0, Expression *Exp1){
        exp0 = Exp0;
        exp1 = Exp1;
    }

    float eval(){
        float e0, e1;
        e0 = exp0->eval();
        e1 = exp1->eval();
        return (e0 > e1)? e0 : e1;
    }
};


/* ############################################################################################ */

enum OPERATOR {
    NONE,
    UNINITED,
    ADD,
    SUB,
    MUL,
    DIV
};

enum FUNCTION {
    VARIABLE,
    COS,
    SIN,
    TAN,
    EXP,
    LOG,
    POW,
    SQRT,
    ABS,
    ANGLE,
    FLOOR,
    CEIL,
    MIN,
    MAX
};


bool parseSubExpression(Expression *&expression, Tokenizer &tok){
    char str[512];
    OPERATOR oper = UNINITED;
    OPERATOR addsubOper = NONE;
    FUNCTION func = VARIABLE;

    Expression *muldiv = NULL;
    Expression *addsub = NULL;
    Expression *temp, *temp2;

    while (tok.next(str)){
        if (tok.isNumeric(str[0]) || str[0] == '('){
            if (oper == UNINITED){
                if (str[0] == '('){
                    parseSubExpression(muldiv, tok);
                } else {
                    muldiv = new Num((float) atof(str));
                }
            } else {
                if (muldiv == NULL){
                    if (addsub != NULL) delete addsub;
                    expression = NULL;
                    return false;
                }

                if (str[0] == '('){
                    parseSubExpression(temp, tok);
                } else {
                    temp = new Num((float) atof(str));
                }

                switch(oper){
                    case MUL:
                        muldiv = new Mul(muldiv, temp);
                        break;
                    case DIV:
                        muldiv = new Div(muldiv, temp);
                        break;
                }
            }
            oper = MUL;
        } else if (tok.isAlphabetical(str[0])){
            if (stricmp(str, "cos") == 0){
                func = COS;
            } else if (stricmp(str, "sin") == 0){
                func = SIN;
            } else if (stricmp(str, "tan") == 0){
                func = TAN;
            } else if (stricmp(str, "exp") == 0){
                func = EXP;
            } else if (stricmp(str, "log") == 0){
                func = LOG;
            } else if (stricmp(str, "pow") == 0){
                func = POW;
            } else if (stricmp(str, "sqrt") == 0){
                func = SQRT;
            } else if (stricmp(str, "abs") == 0){
                func = ABS;
            } else if (stricmp(str, "angle") == 0){
                func = ANGLE;
            } else if (stricmp(str, "floor") == 0){
                func = FLOOR;
            } else if (stricmp(str, "ceil") == 0){
                func = CEIL;
            } else if (stricmp(str, "min") == 0){
                func = MIN;
            } else if (stricmp(str, "max") == 0){
                func = MAX;
            } else {
                // Variable
                func = VARIABLE;
            }

            tok.next(str);
            if (str[0] != '('){
                if (muldiv != NULL) delete muldiv;
                if (addsub != NULL) delete addsub;
                expression = NULL;
                return false;
            } else {
                parseSubExpression(temp, tok);

                switch(func){
                    case COS:
                        temp = new Cos(temp);
                        break;
                    case SIN:
                        temp = new Sin(temp);
                        break;
                    case TAN:
                        temp = new Tan(temp);
                        break;
                    case EXP:
                        temp = new Exp(temp);
                        break;
                    case LOG:
                        temp = new Log(temp);
                        break;
                    case POW:
                        parseSubExpression(temp2, tok);
                        temp = new Pow(temp, temp2);
                        break;
                    case SQRT:
                        temp = new Sqrt(temp);
                        break;
                    case ABS:
                        temp = new Abs(temp);
                        break;
                    case ANGLE:
                        parseSubExpression(temp2, tok);
                        temp = new Angle(temp, temp2);
                        break;
                    case FLOOR:
                        temp = new Floor(temp);
                        break;
                    case CEIL:
                        temp = new Ceil(temp);
                        break;
                    case MIN:
                        parseSubExpression(temp2, tok);
                        temp = new Min(temp, temp2);
                        break;
                    case MAX:
                        parseSubExpression(temp2, tok);
                        temp = new Max(temp, temp2);
                        break;
                }

                if (oper == UNINITED){
                    muldiv = temp;
                } else {
                    switch(oper){
                        case MUL:
                            muldiv = new Mul(muldiv, temp);
                            break;
                        case DIV:
                            muldiv = new Div(muldiv, temp);
                            break;
                    }
                }
            }
            oper = MUL;

        } else {
            if (str[0] == ')' || str[0] == ',') break;

            switch(str[0]){
                case '*':
                    oper = MUL;
                    break;
                case '/':
                    oper = DIV;
                    break;
                case '+':
                case '-':
                    oper = UNINITED;
                    if (addsubOper == NONE){
                        addsub = muldiv;
                        muldiv = NULL;
                    } else {
                        if (addsubOper == ADD){
                            addsub = new Add(addsub, muldiv);
                        } else {
                            addsub = new Sub(addsub, muldiv);
                        }
                        muldiv = NULL;
                    }
                    addsubOper = str[0] == '+'? ADD : SUB;
                    break;
                default:
                    // Free:a minne
                    return false;
            }

        }
    }

    if (addsub != NULL){
        if (muldiv != NULL){
            if (addsubOper == ADD){
                expression = new Add(addsub, muldiv);
            } else {
                expression = new Sub(addsub, muldiv);
            }
        } else {
            delete addsub;
            expression = NULL;
            return false;
        }
    } else if (muldiv != NULL){
        expression = muldiv;
    } else {
        delete muldiv;
        expression = NULL;
        return false;
    }
    return true;
}

bool parseExpression(Expression *&expression, char *expressionString){
    Tokenizer tok;
    tok.setString(expressionString);

    expression = NULL;

    return (parseSubExpression(expression, tok) && expression != NULL);
}





#endif // _EXPRESSION_H_