#ifndef _GLIZER_H_
#define _GLIZER_H_

#include "3D/Vertex.h"
#include "3D/Color.h"


class GLizer {
protected:
	Vertex currentNormal;
	Color currentColor;
	float currentTexCoords[32];

	DWORD fvf;
	D3DPRIMITIVETYPE type;

	unsigned char *fvfBuffer;
	unsigned int bufferSize;
	unsigned int nVertices;
	unsigned int vertexSize;

	bool normals;
	bool colors;
	int nTextures;
public:

	GLizer();
	~GLizer();

	void glBegin(D3DPRIMITIVETYPE Type, bool Normals, bool Colors, int NTextures);
	void glEnd();

	void glVertex(Vertex &vertex);
	void glVertex(float x, float y, float z);
	void glNormal(Vertex &normal);
	void glColor(Color &color);
	void glTexCoord(float s, float t, float r, float q);
	void glMultiTexCoord(int textureUnit, float s, float t, float r, float q);
};

GLizer::GLizer(){
	currentNormal = Vertex(0,0,1);
	currentColor  = Color(1,1,1,1);
	ZeroMemory(currentTexCoords, sizeof(currentTexCoords));

	fvfBuffer = (unsigned char *) malloc(bufferSize = 64);
}

GLizer::~GLizer(){
	if (fvfBuffer != NULL) free(fvfBuffer);
}

void GLizer::glBegin(D3DPRIMITIVETYPE Type, bool Normals, bool Colors, int NTextures){
	nVertices = 0;
	normals = Normals;
	colors = Colors;
	nTextures = NTextures;


	fvf = D3DFVF_XYZ;
	vertexSize = 12;
	if (normals){
		fvf |= D3DFVF_NORMAL;
		vertexSize += 12;
	}
	if (colors){
		fvf |= D3DFVF_DIFFUSE;
		vertexSize += 4;
	}

	for (int i = 0; i < nTextures; i++){
		fvf |= D3DFVF_TEXCOORDSIZE4(i);
	}

	fvf |= nTextures << D3DFVF_TEXCOUNT_SHIFT;

	vertexSize += nTextures * 4 * sizeof(float);

	type = Type;
}

void GLizer::glEnd(){
    d3ddevice->SetVertexShader(fvf);

	int nPrimitives;
	switch (type){
	case D3DPT_POINTLIST:
		nPrimitives = nVertices;
		break;
    case D3DPT_LINELIST:
		nPrimitives = nVertices / 2;
		break;
    case D3DPT_LINESTRIP:
		nPrimitives = nVertices - 1;
		break;
    case D3DPT_TRIANGLELIST:
		nPrimitives = nVertices / 3;
		break;
    case D3DPT_TRIANGLESTRIP:
    case D3DPT_TRIANGLEFAN:
		nPrimitives = nVertices - 2;
		break;
	default:
		nPrimitives = 0;
	}
	d3ddevice->DrawPrimitiveUP(type, nPrimitives, fvfBuffer, vertexSize);
}

void GLizer::glVertex(Vertex &vertex){
	if ((nVertices + 1) * vertexSize > bufferSize){
		fvfBuffer = (unsigned char *) realloc(fvfBuffer, bufferSize = (nVertices + 1) * vertexSize);
	}
	unsigned char *dest = fvfBuffer + nVertices * vertexSize;

	*((Vertex *) dest) = vertex;
	dest += 12;

	if (normals){
		*((Vertex *) dest) = currentNormal;
		dest += 12;
	}

	if (colors){
		*((unsigned int *) dest) = currentColor.getRGBA();
		dest += 4;
	}

	if (nTextures > 0){
		memcpy(dest, currentTexCoords, nTextures * 4 * sizeof(float));
		dest += nTextures * 4 * sizeof(float);
	}


	nVertices++;
}

void GLizer::glVertex(float x, float y = 0.0f, float z = 0.0f){
	glVertex(Vertex(x,y,z));
}


void GLizer::glNormal(Vertex &normal){
	currentNormal = normal;
}

void GLizer::glColor(Color &color){
	currentColor = color;
}

void GLizer::glTexCoord(float s, float t = 0, float r = 0, float q = 1){
	currentTexCoords[0] = s;
	currentTexCoords[1] = t;
	currentTexCoords[2] = r;
	currentTexCoords[3] = q;
}

void GLizer::glMultiTexCoord(int textureUnit, float s, float t = 0, float r = 0, float q = 1){
	currentTexCoords[textureUnit * 4    ] = s;
	currentTexCoords[textureUnit * 4 + 1] = t;
	currentTexCoords[textureUnit * 4 + 2] = r;
	currentTexCoords[textureUnit * 4 + 3] = q;
}

#endif // _GLIZER_H_