                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _TEXTURE_H_
#define _TEXTURE_H_

typedef int TextureID;

#define TEXTURE_DISABLED (-1)

enum TexEnv {
	TEXENV_INVALID,
	TEXENV_MODULATE,
    TEXENV_MODULATE_PRIMARY,
	TEXENV_REPLACE,
	TEXENV_ADD,
    TEXENV_INTERPOLATE,
	TEXENV_DOT3,
	TEXENV_DOT3_NEXT,
};

#define TEXFLAGS_RENORMALIZE 0x1

#include "../Imaging/Image.h"
#include "../3D/TextureUtils.h"
#include "Shader.h"
#include "Common/Set.h"

/* ################################################################################################################### */

class Texture {
protected:
	LPDIRECT3DBASETEXTURE8 textureID;

	D3DTEXTUREFILTERTYPE minFilter, magFilter, mipFilter;
	D3DTEXTUREADDRESS wrapModeS, wrapModeT, wrapModeR;
	D3DFORMAT internalFormat;

	unsigned int width, height, depth, channels;

public:
	Texture();
	virtual ~Texture();

	void setFilter(D3DTEXTUREFILTERTYPE minfilter, D3DTEXTUREFILTERTYPE magfilter, D3DTEXTUREFILTERTYPE mipfilter);
	void setWrapMode(D3DTEXTUREADDRESS wrapS, D3DTEXTUREADDRESS wrapT, D3DTEXTUREADDRESS wrapR);
    void setInternalFormat(D3DFORMAT iFormat);
	virtual bool upload() = 0;
	virtual bool unload();
	virtual bool activate(int textureUnit);
	virtual bool deactivate(int textureUnit);
};

/* ----------------------------------------------------*/

Texture::Texture(){
	width = height = depth = channels = 0;
	textureID = NULL;
	minFilter = D3DTEXF_LINEAR;
	magFilter = D3DTEXF_LINEAR;
	mipFilter = D3DTEXF_LINEAR;
	wrapModeS = D3DTADDRESS_WRAP;
	wrapModeT = D3DTADDRESS_WRAP;
	wrapModeR = D3DTADDRESS_WRAP;
	internalFormat = D3DFMT_A8R8G8B8;
}

Texture::~Texture(){

}

void Texture::setFilter(D3DTEXTUREFILTERTYPE minfilter, D3DTEXTUREFILTERTYPE magfilter, D3DTEXTUREFILTERTYPE mipfilter){
	minFilter = minfilter;
	magFilter = magfilter;
	mipFilter = mipfilter;
}

void Texture::setWrapMode(D3DTEXTUREADDRESS wrapS, D3DTEXTUREADDRESS wrapT, D3DTEXTUREADDRESS wrapR = D3DTADDRESS_WRAP){
	wrapModeS = wrapS;
	wrapModeT = wrapT;
    wrapModeR = wrapR;
}

void Texture::setInternalFormat(D3DFORMAT iFormat){
    internalFormat = iFormat;
}

bool Texture::unload(){
	if (textureID != NULL){
		textureID->Release();
		textureID = NULL;
	}

	return true;
}

bool Texture::activate(int textureUnit){
	d3ddevice->SetTexture(textureUnit, textureID);

	d3ddevice->SetTextureStageState(textureUnit, D3DTSS_MINFILTER, minFilter);
	d3ddevice->SetTextureStageState(textureUnit, D3DTSS_MAGFILTER, magFilter);
	d3ddevice->SetTextureStageState(textureUnit, D3DTSS_MIPFILTER, mipFilter);

	d3ddevice->SetTextureStageState(textureUnit, D3DTSS_ADDRESSU,  wrapModeS);
	d3ddevice->SetTextureStageState(textureUnit, D3DTSS_ADDRESSV,  wrapModeT);
//	d3ddevice->SetTextureStageState(textureUnit, D3DTSS_ADDRESSW,  wrapModeR);

	return true;
}

bool Texture::deactivate(int textureUnit){
	d3ddevice->SetTexture(textureUnit, NULL);
	return true;
}


#include "Texture2D.h"
#include "CubeMap.h"
#include "ProceduralTexture.h"

/* ################################################################################################################### */

class TextureHandler {
private:
	Set <Texture *> textures;
	int nSimultaneousTextures;
    Set <FragmentShader *> fragmentShaders;
    Set <VertexShader   *> vertexShaders;

	TextureID *currentTextures;
	TextureID *selectedTextures;
	
	TexEnv *currentTexEnvs;
	TexEnv *selectedTexEnvs;

    ShaderID selectedVertexShader,   currentVertexShader;
    ShaderID selectedFragmentShader, currentFragmentShader;

	D3DBLEND selectedSrcFactor, selectedDestFactor;
	D3DBLEND currentSrcFactor, currentDestFactor;
	D3DBLENDOP currentBlendMode, selectedBlendMode;
	bool alphaBlendingEnabled;

	void reset();
	
public:
	TextureHandler();
	~TextureHandler();

	void setActiveTexture(int textureUnit);

	TextureID loadFromFile(const char *fileName);
    TextureID loadCubeMapFromFiles(const char *filePosX, const char *fileNegX, const char *filePosY, const char *fileNegY, const char *filePosZ, const char *fileNegZ);
    TextureID createNormalizingCubeMap(unsigned int size);

	TextureID loadFromMemory(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels);
	TextureID createProceduralTexture(PROCEDURALTEXTUREFUNCTION proceduralFunc);
	bool update(TextureID texture, __int64 timeStamp);

	bool toNormalMap(TextureID texture);
	
//    bool setTextures(TextureID t0, TextureID t1, TextureID t2, TextureID t3, TextureID t4, TextureID t5, TextureID t6, TextureID t7, ...);

    bool setTextures(TextureID t0);
    bool setTextures(TextureID t0, TextureID t1);
    bool setTextures(TextureID t0, TextureID t1, TextureID t2);
    bool setTextures(TextureID t0, TextureID t1, TextureID t2, TextureID t3);

	//bool setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2, TexEnv t3, TexEnv t4, TexEnv t5, TexEnv t6, TexEnv t7, ...);
    bool setTexEnvs(TexEnv t0);
    bool setTexEnvs(TexEnv t0, TexEnv t1);
    bool setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2);
    bool setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2, TexEnv t3);

	bool setBlending(D3DBLEND srcFactor, D3DBLEND destFactor);
	bool changeBlending(D3DBLEND srcFactor, D3DBLEND destFactor);
	bool setBlendMode(D3DBLENDOP mode);
	bool apply();

	//ShaderID createVertexShader(const char *shaderCode);
    bool setVertexShader(ShaderID shader);

	ShaderID createFragmentShader(const char *shaderCode);
    bool setFragmentShader(ShaderID shader);

	bool upload(TextureID texture);
	bool uploadAll();
	bool unload(TextureID texture);
	bool unloadAll();

	void setFilter(TextureID texture, D3DTEXTUREFILTERTYPE minfilter, D3DTEXTUREFILTERTYPE magfilter, D3DTEXTUREFILTERTYPE mipfilter);
	void setWrapMode(TextureID texture, D3DTEXTUREADDRESS wrapModeS, D3DTEXTUREADDRESS wrapModeT);
    void setInternalFormat(TextureID texture, D3DFORMAT internalFormat);

};

/* ----------------------------------------------------*/

void TextureHandler::reset(){
	for (int i = 0; i < nSimultaneousTextures; i++){
		selectedTextures [i] = TEXTURE_DISABLED;
		selectedTexEnvs  [i] = TEXENV_MODULATE;
	}

	selectedSrcFactor  = D3DBLEND_ONE;
	selectedDestFactor = D3DBLEND_ZERO;
	selectedBlendMode  = D3DBLENDOP_ADD;

    selectedVertexShader   = SHADER_DISABLED;
    selectedFragmentShader = SHADER_DISABLED;
}


/* ----------------------------------------------------*/

TextureHandler::TextureHandler(){
	nSimultaneousTextures = caps.MaxSimultaneousTextures;

	currentTextures  = new TextureID [nSimultaneousTextures];
	selectedTextures = new TextureID [nSimultaneousTextures];

	selectedTexEnvs  = new TexEnv    [nSimultaneousTextures];
	currentTexEnvs   = new TexEnv    [nSimultaneousTextures];

	reset();

	for (int i = 0; i < nSimultaneousTextures; i++){
		currentTextures [i] = TEXTURE_DISABLED;
		currentTexEnvs  [i] = TEXENV_INVALID;
	}

	currentSrcFactor  = D3DBLEND_ONE;
	currentDestFactor = D3DBLEND_ZERO;
	currentBlendMode  = D3DBLENDOP_ADD;
	alphaBlendingEnabled = false;
}

TextureHandler::~TextureHandler(){
	int i, len = textures.getSize();
	for (i = 0; i < len; i++){
		if (textures[i] != NULL){
			delete textures[i];
		}
	}

    len = fragmentShaders.getSize();
	for (i = 0; i < len; i++){
		if (fragmentShaders[i] != NULL){
			delete fragmentShaders[i];
		}
	}

    len = vertexShaders.getSize();
	for (i = 0; i < len; i++){
		if (vertexShaders[i] != NULL){
			delete vertexShaders[i];
		}
	}

	delete selectedTexEnvs;
	delete currentTexEnvs;

	delete selectedTextures;
	delete currentTextures;
}

TextureID TextureHandler::loadFromFile(const char *fileName){
	Texture2D *texture = new Texture2D();

	if (texture->loadFromFile(fileName)){
		int i, len = textures.getSize();

		for (i = 0; i < len; i++){
			if (textures[i] == NULL){
				textures[i] = texture;
				return i;
			}
		}
		return textures.add(texture);
	} else {
		delete texture;
		return TEXTURE_DISABLED;
	}	
}


TextureID TextureHandler::loadCubeMapFromFiles(const char *filePosX, const char *fileNegX, const char *filePosY, const char *fileNegY, const char *filePosZ, const char *fileNegZ){
	CubeMap *cubemap = new CubeMap();

	if (cubemap->loadFromFiles(filePosX, fileNegX, filePosY, fileNegY, filePosZ, fileNegZ)){
		int i, len = textures.getSize();

		for (i = 0; i < len; i++){
			if (textures[i] == NULL){
				textures[i] = cubemap;
				return i;
			}
		}
		return textures.add(cubemap);
	} else {
		delete cubemap;
		return TEXTURE_DISABLED;
	}	
}

TextureID TextureHandler::createNormalizingCubeMap(unsigned int size){
	CubeMap *cubemap = new CubeMap();

    cubemap->loadNormalizing(size);

    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
			textures[i] = cubemap;
            return i;
        }
    }
    return textures.add(cubemap);
}

TextureID TextureHandler::loadFromMemory(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels){
	Texture2D *texture = new Texture2D();

	texture->loadFromMemory(pixels, width, height, channels);
    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
            textures[i] = texture;
            return i;
        }
    }
	return textures.add(texture);
}

TextureID TextureHandler::createProceduralTexture(PROCEDURALTEXTUREFUNCTION proceduralFunc){
	ProceduralTexture *texture = new ProceduralTexture(proceduralFunc);

    int i, len = textures.getSize();
    for (i = 0; i < len; i++){
        if (textures[i] == NULL){
            textures[i] = texture;
            return i;
        }
    }
	return textures.add(texture);
}

bool TextureHandler::update(TextureID texture, __int64 timeStamp){
    return ((ProceduralTexture *) textures[texture])->update(timeStamp);
}

bool TextureHandler::toNormalMap(TextureID texture){
	return ((Texture2D *) textures[texture])->toNormalMap();
}


bool TextureHandler::setTextures(TextureID t0){
    selectedTextures[0] = t0;
    return true;
}

bool TextureHandler::setTextures(TextureID t0, TextureID t1){
    selectedTextures[0] = t0;
    selectedTextures[1] = t1;
    return true;
}

bool TextureHandler::setTextures(TextureID t0, TextureID t1, TextureID t2){
    selectedTextures[0] = t0;
    selectedTextures[1] = t1;
    selectedTextures[2] = t2;
    return true;
}

bool TextureHandler::setTextures(TextureID t0, TextureID t1, TextureID t2, TextureID t3){
    selectedTextures[0] = t0;
    selectedTextures[1] = t1;
    selectedTextures[2] = t2;
    selectedTextures[3] = t3;
    return true;
}

bool TextureHandler::setTexEnvs(TexEnv t0){
	selectedTexEnvs[0] = t0;
    return true;
}

bool TextureHandler::setTexEnvs(TexEnv t0, TexEnv t1){
	selectedTexEnvs[0] = t0;
	selectedTexEnvs[1] = t1;
    return true;
}

bool TextureHandler::setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2){
	selectedTexEnvs[0] = t0;
	selectedTexEnvs[1] = t1;
	selectedTexEnvs[2] = t2;
    return true;
}

bool TextureHandler::setTexEnvs(TexEnv t0, TexEnv t1, TexEnv t2, TexEnv t3){
	selectedTexEnvs[0] = t0;
	selectedTexEnvs[1] = t1;
	selectedTexEnvs[2] = t2;
	selectedTexEnvs[3] = t3;
    return true;
}

bool TextureHandler::setBlending(D3DBLEND srcFactor, D3DBLEND destFactor){
    selectedSrcFactor  = srcFactor;
    selectedDestFactor = destFactor;
	return true;
}

bool TextureHandler::changeBlending(D3DBLEND srcFactor, D3DBLEND destFactor){
	if (srcFactor != D3DBLEND_ONE || destFactor != D3DBLEND_ZERO){
		if (srcFactor != currentSrcFactor){
			d3ddevice->SetRenderState(D3DRS_SRCBLEND, srcFactor);
			currentSrcFactor = srcFactor;
		}
		
		if (destFactor != currentDestFactor){
			d3ddevice->SetRenderState(D3DRS_DESTBLEND, destFactor);
			currentDestFactor = destFactor;
		}

		if (!alphaBlendingEnabled){
			d3ddevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
			alphaBlendingEnabled = true;
		}
	} else {
		if (alphaBlendingEnabled){
			d3ddevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
			alphaBlendingEnabled = false;
		}
	}

    return true;
}

bool TextureHandler::setBlendMode(D3DBLENDOP mode){
	selectedBlendMode = mode;
	return true;
}


bool TextureHandler::apply(){
    // Setup texture binding and texture enviroments
	for (int i = 0; i < nSimultaneousTextures; i++){
		if (selectedTextures[i] == TEXTURE_DISABLED){
			if (currentTextures[i] != TEXTURE_DISABLED){
                // Going from textured to non-textured
				textures[currentTextures[i]]->deactivate(i);
				currentTextures[i] = TEXTURE_DISABLED;
			}
		} else {

			if (currentTextures[i] == TEXTURE_DISABLED){
				// Going from non-textured to textured
				textures[selectedTextures[i]]->activate(i);
				currentTextures[i] = selectedTextures[i];
			} else {
				if (currentTextures[i] != selectedTextures[i]){
					// Going from one texture to another texture
					textures[currentTextures[i]]->deactivate(i);
					textures[selectedTextures[i]]->activate(i);

					currentTextures[i] = selectedTextures[i];
                }
			}
		}


		if (currentTexEnvs[i] != selectedTexEnvs[i]){
			switch(selectedTexEnvs[i]){
			case TEXENV_MODULATE:
				d3ddevice->SetTextureStageState(i, D3DTSS_COLOROP,   D3DTOP_MODULATE);
				d3ddevice->SetTextureStageState(i, D3DTSS_COLORARG1, D3DTA_CURRENT);
				d3ddevice->SetTextureStageState(i, D3DTSS_COLORARG2, D3DTA_TEXTURE);

				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);
				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAARG1, D3DTA_CURRENT);
				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);
				break;
/*			case TEXENV_MODULATE_PRIMARY:
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_MODULATE);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_PRIMARY_COLOR_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
				break;
				*/
			case TEXENV_REPLACE:
				d3ddevice->SetTextureStageState(i, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
				d3ddevice->SetTextureStageState(i, D3DTSS_COLORARG1, D3DTA_TEXTURE);

				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
				break;
			case TEXENV_ADD:
				d3ddevice->SetTextureStageState(i, D3DTSS_COLOROP,   D3DTOP_ADD);
				d3ddevice->SetTextureStageState(i, D3DTSS_COLORARG1, D3DTA_CURRENT);
				d3ddevice->SetTextureStageState(i, D3DTSS_COLORARG2, D3DTA_TEXTURE);

				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAOP,   D3DTOP_ADD);
				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAARG1, D3DTA_CURRENT);
				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);
				break;
			/*
			case TEXENV_INTERPOLATE:
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_INTERPOLATE_ARB);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB,  GL_PRIMARY_COLOR_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
				break;
				*/
			case TEXENV_DOT3:
				d3ddevice->SetTextureStageState(i, D3DTSS_COLOROP,   D3DTOP_DOTPRODUCT3);
				d3ddevice->SetTextureStageState(i, D3DTSS_COLORARG1, D3DTA_CURRENT);
				d3ddevice->SetTextureStageState(i, D3DTSS_COLORARG2, D3DTA_TEXTURE);

				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAOP,   D3DTOP_DOTPRODUCT3);
				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAARG1, D3DTA_CURRENT);
				d3ddevice->SetTextureStageState(i, D3DTSS_ALPHAARG2, D3DTA_TEXTURE);
				break;
				/*
			case TEXENV_DOT3_NEXT:
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_DOT3_RGBA_ARB);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_TEXTURE0_ARB + i);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
				
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE0_ARB + i + 1);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 1);
				break;
				*/
			}
			
			
			currentTexEnvs[i] = selectedTexEnvs[i];
		}
	}

    // Setup fragment shaders
    if (selectedFragmentShader != currentFragmentShader){
		if (selectedFragmentShader == SHADER_DISABLED){
			FragmentShader::deactivate();
		} else {
			fragmentShaders[selectedFragmentShader]->activate();
		}
		currentFragmentShader = selectedFragmentShader;
    }

    // Setup blending
	changeBlending(selectedSrcFactor, selectedDestFactor);


    // Setup blend mode
	if (selectedBlendMode != currentBlendMode){
		d3ddevice->SetRenderState(D3DRS_BLENDOP, selectedBlendMode);
		currentBlendMode = selectedBlendMode;
	}

	
	reset();

	return true;
}

/*ShaderID TextureHandler::createVertexShader(const char *shaderCode){
	return vertexShaders.add(new VertexShader(shaderCode));
}*/

bool TextureHandler::setVertexShader(ShaderID shader){
    selectedVertexShader = shader;
    return true;
}

ShaderID TextureHandler::createFragmentShader(const char *shaderCode){
	return fragmentShaders.add(new FragmentShader(shaderCode));
}

bool TextureHandler::setFragmentShader(ShaderID shader){
    selectedFragmentShader = shader;
    return true;
}



bool TextureHandler::upload(TextureID texture){
	return textures[texture]->upload();
}

bool TextureHandler::uploadAll(){
	int i, len = textures.getSize();
	for (i = 0; i < len; i++){
		textures[i]->upload();
	}
	return true;
}

bool TextureHandler::unload(TextureID texture){
	return textures[texture]->unload();
}

bool TextureHandler::unloadAll(){
	int i, len;

	for (i = 0; i < nSimultaneousTextures; i++){
		if (currentTextures[i] != TEXTURE_DISABLED){
			textures[currentTextures[i]]->deactivate(i);
			currentTextures[i] = TEXTURE_DISABLED;
		}
	}

	len = textures.getSize();
	for (i = 0; i < len; i++){
		textures[i]->unload();
	}
	return true;
}

void TextureHandler::setFilter(TextureID texture, D3DTEXTUREFILTERTYPE minfilter, D3DTEXTUREFILTERTYPE magfilter, D3DTEXTUREFILTERTYPE mipfilter){
	textures[texture]->setFilter(minfilter, magfilter, mipfilter);
}

void TextureHandler::setWrapMode(TextureID texture, D3DTEXTUREADDRESS wrapModeS, D3DTEXTUREADDRESS wrapModeT){
	textures[texture]->setWrapMode(wrapModeS, wrapModeT);
}

void TextureHandler::setInternalFormat(TextureID texture, D3DFORMAT internalFormat){
    textures[texture]->setInternalFormat(internalFormat);
}

/* ################################################################################################################### */


#endif // _TEXTURE_H_