                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _TEXTURE2D_H_
#define _TEXTURE2D_H_

/* ################################################################################################################### */


class Texture2D : public Texture {
protected:
	unsigned char *pixels;
	unsigned int format;
public:
	Texture2D();
	virtual ~Texture2D();

	bool loadFromFile(const char *fileName);
	bool loadFromMemory(unsigned char *Pixels, unsigned int Width, unsigned int Height, unsigned int Channels);
//    bool createSpecularMap(unsigned int Width, unsigned int Height, float minExponent, float maxExponent);
	bool toNormalMap();

	bool upload();
};

/* ----------------------------------------------------*/

Texture2D::Texture2D(){
	pixels = NULL;
	width = height = depth = channels = format = 0;


	textureID = NULL;
	minFilter = D3DTEXF_LINEAR;
	magFilter = D3DTEXF_LINEAR;
	mipFilter = D3DTEXF_LINEAR;
	wrapModeS = D3DTADDRESS_WRAP;
	wrapModeT = D3DTADDRESS_WRAP;
	wrapModeR = D3DTADDRESS_WRAP;
	internalFormat = D3DFMT_A8R8G8B8;
}

Texture2D::~Texture2D(){
	if (pixels != NULL) delete pixels;
	if (textureID != NULL) textureID->Release();
}

bool Texture2D::loadFromFile(const char *fileName){
	if (pixels != NULL){
		delete pixels;
		pixels = NULL;
	}
    if (loadImage(fileName, pixels, width, height, channels, format) == IMAGE_SUCCESSFUL){
        depth = 1;
        return true;
    }
    return false;
}

bool Texture2D::loadFromMemory(unsigned char *Pixels, unsigned int Width, unsigned int Height, unsigned int Channels){
	if (pixels != NULL)	delete pixels;

	pixels   = Pixels;
	width    = Width;
	height   = Height;
    depth    = 1;
	channels = Channels;

	return true;
}

/*bool Texture2D::createSpecularMap(unsigned int Width, unsigned int Height, float minExponent, float maxExponent){
	if (pixels != NULL)	delete pixels;

	width    = Width;
	height   = Height;
    depth    = 1;
	channels = 1;
    internalFormat = GL_LUMINANCE;
	wrapModeS = GL_CLAMP_TO_EDGE_EXT;
	wrapModeT = GL_CLAMP_TO_EDGE_EXT;

    pixels = new unsigned char[Width * Height];

    unsigned char *dest = pixels;
    unsigned int x,y;
    float exponent;
    for (y = 0; y < height; y++){
        exponent = minExponent + (y * (maxExponent - minExponent) / (height - 1));
        for (x = 0; x < width; x++){
            *dest++ = (unsigned char) (255 * powf((float) x / (float) (width - 1), exponent));
        }
    }

    return true;
}*/

bool Texture2D::toNormalMap(){
	if (pixels != NULL){
		if (channels == 3){
			toGrayscale(pixels, width, height, channels);
		}
		if (channels == 1){
			filterImage(pixels, width, height, channels, Dot3, 1,1,1,1, EDGE_CLAMP, width, height, 3, NULL);
			return true;
		}
	}
	return false;
}

bool Texture2D::upload(){
	if (pixels == NULL) return false;

	if (textureID != NULL){
		textureID->Release();
		textureID = NULL;
	}


	LPDIRECT3DTEXTURE8 texID;
	upload2DImage(texID, pixels, width, height, channels, format, internalFormat);

	textureID = texID;

	delete pixels;
	pixels = NULL;
	
	return true;
}


#endif // _TEXTURE2D_H_