/*
 *  App programming interface (internal header file).
 *
 *  Copyright (c) L.Patrick
 *
 *  This header file never needs to be included by a user's
 *  program. It is only used internally while compiling the
 *  the App library. A program need only include app.h
 *  and link with the app.lib or libapp.a library.
 */

#ifndef _APPINT_H_
#define _APPINT_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stdarg.h>

#undef TRANSPARENT
#include "app.h"

#undef TRANSPARENT
#include <setjmp.h>
#include <windows.h>
#undef TRANSPARENT
#define TRANSPARENT     argb(0xFF,0xFF,0xFF,0xFF)

/*
 *  Fix a few compiler-specific definitions:
 */
#ifndef _export
	#define _export
#endif
#ifndef FAR
	#define FAR
#endif
#ifndef PASCAL
	#define PASCAL
#endif

/*
 *  Platform-specific extra information:
 */

typedef struct AppExtra
{
	HINSTANCE	this_instance;
} AppExtra;

#define app_extra(app) ((AppExtra*)((app)->extra))


typedef struct WinExtra
{
	HWND		hwnd;
	HPALETTE	winpal;
} WinExtra;

#define win_extra(win) ((WinExtra*)((win)->extra))


typedef struct BitmapExtra
{
	HBITMAP		handle;
	HBITMAP		clipmask;
} BitmapExtra;

#define bitmap_extra(bmap) ((BitmapExtra*)((bmap)->extra))


typedef struct GraphicsExtra
{
	int 		kind;
	HDC 		dc;
	HBRUSH		oldbr;
	HBRUSH		brush;
	HPEN		oldpen;
	HPEN		pen;
	HBITMAP		oldbm;
	HPALETTE	oldpal;
	Colour		bg;
} GraphicsExtra;

#define graphics_extra(g) ((GraphicsExtra*)((g)->extra))


typedef struct SubfontExtra
{
	HBITMAP		clipmask;	/* Windows depth-1 stencil */
} SubfontExtra;

#define subfont_extra(sub) ((SubfontExtra*)((sub)->extra))


typedef struct FontExtra
{
	HFONT		fnt;
} FontExtra;

#define font_extra(f) ((FontExtra*)((f)->extra))


typedef struct FolderExtra
{
	int			first;
	HANDLE		hnd;
    WIN32_FIND_DATA	data;
} FolderExtra;

#define folder_extra(f) ((FolderExtra*)((f)))


/*
 *  Internal definitions:
 */

/* Global read-only variables: */

extern char * app_class_name;

/* Controls: */

void	app_do_draw_controls(Graphics *g, int num, Control **list, int clear);

/* Drawing: */

int 	app_bitmap_fill_rect(Graphics *g, Rect r);
int 	app_bitmap_copy_rect(Graphics *dst, Point dp, Graphics *src, Rect sr);
int 	app_bitmap_draw_utf8(Graphics *g, Point p, char *s, int nbytes);
int 	app_bitmap_draw_line(Graphics *g, Point p1, Point p2);

int 	app_image_fill_rect(Graphics *g, Rect r);
int 	app_image_copy_rect(Graphics *dst, Point dp, Graphics *src, Rect sr);
int 	app_image_draw_utf8(Graphics *g, Point p, char *s, int nbytes);

int 	app_window_fill_rect(Graphics *g, Rect r);
int 	app_window_copy_rect(Graphics *dst, Point dp, Graphics *src, Rect sr);
int 	app_window_draw_utf8(Graphics *g, Point p, char *s, int nbytes);
int 	app_window_draw_line(Graphics *g, Point p1, Point p2);

/* Events: */

long FAR PASCAL
_export app_winproc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

/* Graphics context manipulation: */

Graphics *app_get_window_redraw(HDC dc, Window *w);

/* Image/bitmap conversion: */

HBITMAP app_image_to_clipmask(App *app, Image *img);

/* Palettes: */

int app_realize_palette(Window *win);

/* Folders and file names: */

char * app_to_dos_path(char *filepath);
char * app_from_dos_path(char *filepath);

#endif // _APPINT_H_

