                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/

#ifndef _IMAGEUTILS_H_
#define _IMAGEUTILS_H_

#include <math.h>
#include "../Common/MyMath.h"

#define EDGE_CLAMP 0
#define EDGE_WRAP  1

typedef int (*FILTER_FUNC)(unsigned char *dest, unsigned char *src, int *params);

void filterImage(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels,
				 FILTER_FUNC filter, int filterLeft, int filterTop, int filterRight, int filterBottom, unsigned int edgeMode,
				 unsigned int newWidth, unsigned int newHeight, unsigned int newChannels, int *params = NULL){

	unsigned char *newPixels  = (unsigned char *) malloc(newWidth * newHeight * newChannels);
	unsigned int filterWidth  = filterLeft + filterRight  + 1;
	unsigned int filterHeight = filterTop  + filterBottom + 1;
	unsigned char *dest = newPixels;
	unsigned char *src = (unsigned char *) malloc(filterWidth * filterHeight * channels);
	unsigned char *samplePoint, *srcFill;
	int x,y,i,j,si,sj,numPixels;
	unsigned int k;


	for (y = 0; y < (signed) height; y++){
		for (x = 0; x < (signed) width; x++){
			srcFill = src;

			switch(edgeMode){
			case EDGE_CLAMP:
				for (j = y - filterTop; j <= y + filterBottom; j++){
					sj = (j < 0)? 0:(j >= (signed) height? height - 1:j);
					for (i = x - filterLeft; i <= x + filterRight; i++){
						si = (i < 0)? 0:(i >= (signed) width? width - 1:i);
						
						samplePoint = pixels + channels * (sj * width + si);
						
						for (k = 0; k < channels; k++){
							*srcFill++ = *samplePoint++;
						}
					}
				}
				break;
			case EDGE_WRAP:
				for (j = y - filterTop; j <= y + filterBottom; j++){
					sj = (j < 0)? (j + height):(j >= (signed) height? j - height:j);
					for (i = x - filterLeft; i <= x + filterRight; i++){
						si = (i < 0)? (i + width):(i >= (signed) width? i - width:i);
						
						samplePoint = pixels + channels * (sj * width + si);
						
						for (k = 0; k < channels; k++){
							*srcFill++ = *samplePoint++;
						}
					}
				}
				break;
			}
						
			
			numPixels = filter(dest, src, params);

			dest += newChannels * numPixels;
		}
	}


	free(src);
	free(pixels);
	pixels = newPixels;
	width = newWidth;
	height = newHeight;
	channels = newChannels;
}



// Some useful standard filters

// SobelX detects vertical edges
int SobelX(unsigned char *dest, unsigned char *src, int *params){
	int j,k;

	for (j = 0; j < 3; j++){
		k = src[6 + j] - src[j] + ((src[15 + j] - src[9 + j]) << 1) + src[24 + j] - src[18 + j];
		dest[j] = (k < 0)? 0 : (k > 255)? 255 : k;
	}
	return 1;
}

// SobelY detects horizontal edges
int SobelY(unsigned char *dest, unsigned char *src, int *params){
	int j,k;

	for (j = 0; j < 3; j++){
		k = src[j] + src[6 + j] + ((src[3 + j] - src[21 + j]) << 1) - src[18 + j] - src[24 + j];
		dest[j] = (k < 0)? 0 : (k > 255)? 255 : k;
	}
	return 1;
}

// SobelXY detects all edges
int SobelXY(unsigned char *dest, unsigned char *src, int *params){
	int x,y,j,k;

	for (j = 0; j < 3; j++){
		x = src[6 + j] - src[j] + ((src[15 + j] - src[9 + j]) << 1) + src[24 + j] - src[18 + j];
		y = src[j] + src[6 + j] + ((src[3 + j] - src[21 + j]) << 1) - src[18 + j] - src[24 + j];
		k = (int) sqrtf(float(x * x + y * y));
		dest[j] = (k < 255)? k : 255;
	}
	    
	return 1;
}

// Dot3 creates a normal map from a heightmap
int Dot3(unsigned char *dest, unsigned char *src, int *params){
	int x,y,len;

	x = src[0] - src[2] + ((src[3] - src[5]) << 1) + src[6] - src[8];
	y = src[0] + src[2] + ((src[1] - src[7]) << 1) - src[6] - src[8];

	len = int(1044480.0f / sqrtf(float(x * x + y * y + 65536)));
	x *= len;
	y *= len;

	dest[0] = (unsigned char) ((x + 1044480) >> 13);
	dest[1] = (unsigned char) ((y + 1044480) >> 13);
	dest[2] = (unsigned char) ((len + 4080) >> 5);

	return 1;
}

// Laplace detects all edges, but responds on the second derivate
int Laplace(unsigned char *dest, unsigned char *src, int *params){
	for (int j = 0; j < 3; j++){
		dest[j] = abs(src[j + 3] + src[j+9] - (src[j + 12] << 2) + src[j + 15] + src[j + 21]);
	}
	return 1;
}


int Median(unsigned char *dest, unsigned char *src, int *params){
	unsigned char tmp;
	int i, k;
	bool changed;

	for (k = 0; k < 3; k++){
		do {
			changed = false;
			for (i = k; i < 24; i += 3){
				if (src[i] > src[i + 3]){
					tmp = src[i];
					src[i] = src[i + 3];
					src[i + 3] = tmp;
					changed = true;
				}
			}	
		} while (changed);
		dest[k] = src[12 + k];
	}
	return 1;
}

int Soften(unsigned char *dest, unsigned char *src, int *params){
	for (int k = 0; k < 3; k++){
		*dest++ = (*src + 2 * src[3] + src[6] + 2 * src[9] + 4 * src[12] + 2 * src[15] + src[18] + 2 * src[21] + src[24]) >> 4;
		src++;
	}
	return 1;
}

int Blur(unsigned char *dest, unsigned char *src, int *params){
	for (int k = 0; k < 3; k++){
		*dest++ = (*src + src[3] + src[6] + src[9] + src[15] + src[18] + src[21] + src[24]) >> 3;
		src++;
	}
	return 1;
}

int Sharpen(unsigned char *dest, unsigned char *src, int *params){
	int mean, val;
	for (int k = 0; k < 3; k++){
		mean = (*src + src[3] + src[6] + src[9] + src[15] + src[18] + src[21] + src[24]) >> 3;

		val = ((src[12] - mean) << 1) + mean;

		*dest++ = (val < 0)? 0:(val > 255)? 255:val;

		src++;
	}
	return 1;
}

// ##############################################################################################################################

// Basic image operations

unsigned int invert(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels){
	__asm {
		MOV		EDI, pixels

		MOV		EAX, width
		IMUL	EAX, height
		IMUL	EAX, channels
		JMP		TheLoopFirst

		TheLoop128:
			NOT	DWORD PTR [EDI]
			NOT	DWORD PTR [EDI+4]
			NOT	DWORD PTR [EDI+8]
			NOT	DWORD PTR [EDI+12]
			ADD	EDI, 16
		TheLoopFirst:
			SUB	EAX, 16
		JGE	TheLoop128
		ADD	EAX, 16
		JZ	End	
		TheLoop8:
			NOT	BYTE PTR [EDI]
			INC	EDI
			DEC	EAX
		JNZ	TheLoop8
		
		End:
	}
	return IMAGE_SUCCESSFUL;
}

unsigned int add(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels){
	unsigned int b, len = width * height * channels;

	do {
		b = *destPixels + *srcPixels;
		*destPixels++ = (b > 255)? 255:b;
		*srcPixels++;
	} while (--len);
	return IMAGE_SUCCESSFUL;
}

unsigned int add_MMX(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels){
	__asm {
		MOV		EDI, destPixels
		MOV		ESI, srcPixels

		MOV		EAX, width
		IMUL	EAX, height
		IMUL	EAX, channels
		JMP		TheLoopFirst

		TheLoop64:
			MOVQ	MM0, [EDI]
			PADDUSB	MM0, [ESI]
			MOVQ	[EDI], MM0
			ADD		EDI, 8
			ADD		ESI, 8
		TheLoopFirst:
			SUB		EAX, 8
		JGE	TheLoop64
		ADD	EAX, 8
		JZ	End	
		MOV	EDX, 255
		TheLoop8:
			MOVZX	EBX, [EDI]
			MOVZX	ECX, [ESI]
			ADD		EBX, ECX
			CMP		EBX, EDX
			CMOVA	EBX, EDX
			MOV		[EDI], BL
			INC		EDI
			INC		ESI
			DEC		EAX
		JNZ	TheLoop8

		End:

		EMMS
	}

	return IMAGE_SUCCESSFUL;
}

unsigned int sub(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels){
	unsigned int len = width * height * channels;
	int b;

	do {
		b = *destPixels - *srcPixels;
		*destPixels++ = (b < 0)? 0:b;
		*srcPixels++;
	} while (--len);
	return IMAGE_SUCCESSFUL;
}

unsigned int sub_MMX(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels){
	__asm {
		MOV		EDI, destPixels
		MOV		ESI, srcPixels

		MOV		EAX, width
		IMUL	EAX, height
		IMUL	EAX, channels
		JMP		TheLoopFirst

		TheLoop64:
			MOVQ	MM0, [EDI]
			PSUBUSB	MM0, [ESI]
			MOVQ	[EDI], MM0
			ADD		EDI, 8
			ADD		ESI, 8
		TheLoopFirst:
			SUB		EAX, 8
		JGE	TheLoop64
		ADD	EAX, 8
		JZ	End	
		XOR	EDX, EDX
		TheLoop8:
			MOVZX	EBX, [EDI]
			MOVZX	ECX, [ESI]
			SUB		EBX, ECX
			CMOVB	EBX, EDX
			MOV		[EDI], BL
			INC		EDI
			INC		ESI
			DEC		EAX
		JNZ	TheLoop8

		End:

		EMMS
	}

	return IMAGE_SUCCESSFUL;
}

unsigned int modulate(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels){
	unsigned int len = width * height * channels;

	do {
		*destPixels++ = (*destPixels * *srcPixels++) >> 8;
	} while (--len);
	return IMAGE_SUCCESSFUL;
}

unsigned int modulate_MMX(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels){
	__asm {
		MOV		EDI, destPixels
		MOV		ESI, srcPixels

		MOV		EAX, width
		IMUL	EAX, height
		IMUL	EAX, channels
		PXOR	MM7, MM7
		JMP		TheLoopFirst

		TheLoop64:
			MOVD		MM0, [EDI]
			MOVD		MM1, [EDI+4]
			MOVD		MM2, [ESI]
			MOVD		MM3, [ESI+4]
			PUNPCKLBW	MM0, MM7
			PUNPCKLBW	MM1, MM7
			PUNPCKLBW	MM2, MM7
			PUNPCKLBW	MM3, MM7

			PMULLW		MM0, MM2
			PMULLW		MM1, MM3
			PSRLW		MM0, 8
			PSRLW		MM1, 8

			PACKUSWB	MM0, MM1
			MOVQ		[EDI], MM0
			ADD		EDI, 8
			ADD		ESI, 8
		TheLoopFirst:
			SUB		EAX, 8
		JGE	TheLoop64
		ADD	EAX, 8
		JZ	End	
		TheLoop8:
			MOVZX	EBX, [EDI]
			MOVZX	ECX, [ESI]
			IMUL	EBX, ECX
			MOV		[EDI], BH
			INC		EDI
			INC		ESI
			DEC		EAX
		JNZ	TheLoop8

		End:

		EMMS
	}

	return IMAGE_SUCCESSFUL;
}

unsigned int alphaBlend(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels, unsigned char alpha){
	unsigned int len = width * height * channels;

	do {
		*destPixels++ = ((255 - alpha) * (*destPixels) + alpha * (*srcPixels++)) >> 8;
	} while (--len);
	return IMAGE_SUCCESSFUL;
}

unsigned int alphaBlend_MMX(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels, unsigned char alpha){
	__asm {
		MOV		EDI, destPixels
		MOV		ESI, srcPixels

		MOV		EAX, width
		IMUL	EAX, height
		IMUL	EAX, channels
		PXOR	MM7, MM7

		XOR		ECX, ECX
		MOV		CL, alpha
		MOV		CH, alpha
		SHL		ECX, 16
		MOV		CL, alpha
		MOV		CH, alpha

		MOV		EBX, ECX
		NOT		EBX

		MOVD	MM5, EBX
		MOVD	MM6, ECX
		PUNPCKLBW	MM5, MM7
		PUNPCKLBW	MM6, MM7

		JMP		TheLoopFirst

		TheLoop64:
			MOVD		MM0, [EDI]
			MOVD		MM1, [EDI+4]
			MOVD		MM2, [ESI]
			MOVD		MM3, [ESI+4]
			PUNPCKLBW	MM0, MM7
			PUNPCKLBW	MM1, MM7
			PUNPCKLBW	MM2, MM7
			PUNPCKLBW	MM3, MM7

			PMULLW		MM0, MM5
			PMULLW		MM1, MM5
			PMULLW		MM2, MM6
			PMULLW		MM3, MM6

			PADDW		MM0, MM2
			PADDW		MM1, MM3

			PSRLW		MM0, 8
			PSRLW		MM1, 8

			PACKUSWB	MM0, MM1
			MOVQ		[EDI], MM0
			ADD		EDI, 8
			ADD		ESI, 8
		TheLoopFirst:
			SUB		EAX, 8
		JGE	TheLoop64
		ADD	EAX, 8
		JZ	End	
		AND	ECX, 0xFF
		TheLoop8:
			XOR		ECX, 0xFF
			MOVZX	EDX, [EDI]
			IMUL	EDX, ECX
			XOR		ECX, 0xFF
			MOVZX	EBX, [ESI]
			IMUL	EBX, ECX
			ADD		EBX, EDX
			MOV		[EDI], BH
			INC		EDI
			INC		ESI
			DEC		EAX
		JNZ	TheLoop8

		End:

		EMMS
	}

	return IMAGE_SUCCESSFUL;
}

unsigned int combine(unsigned char *destPixels, unsigned char *srcPixels, unsigned int width, unsigned int height, unsigned int channels, int destFactor, int srcFactor, int divisor = 1, int bias = 0){
	if (divisor != 0){
		__asm {
			MOV		EDI, destPixels
			MOV		ESI, srcPixels
				
			MOV		ECX, width
			IMUL	ECX, height
			IMUL	ECX, channels
				

			TheLoop:
				MOVZX	EAX, [EDI]
				IMUL	EAX, destFactor
				MOVZX	EBX, [ESI]
				IMUL	EBX, srcFactor
				ADD		EAX, EBX
				XOR		EDX, EDX
				IDIV	divisor

				XOR		EDX, EDX
				ADD		EAX, bias
				CMOVB	EAX, EDX
				NOT		DL
				CMP		EAX, EDX
				CMOVA	EAX, EDX
				
				MOV		[EDI], AL
				
				INC		EDI
				INC		ESI
				DEC		ECX
			JNZ	TheLoop
		}
		
		return IMAGE_SUCCESSFUL;
	} else return IMAGE_IMPROPER_OPERAND;
}

unsigned int treshold(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels, unsigned char lowerBound, unsigned char upperBound){
	/*unsigned int len = width * height * channels;

	do {
		*pixels++ = 255 * (*pixels >= lowerBound && *pixels <= upperBound);
	} while (--len);
	*/
	__asm {
		MOV		EDI, pixels
			
		MOV		ECX, width
		IMUL	ECX, height
		IMUL	ECX, channels

		MOV		BL, lowerBound
		MOV		BH, upperBound
			
			
		TheLoop:
			MOV		AL, [EDI]
			CMP		AL, BL
			SETNAE	DL
			CMP		AL, BH
			SETA	DH
			OR		DL, DH
			DEC		DL			
			MOV		[EDI], DL
			
			INC		EDI
			DEC		ECX
		JNZ	TheLoop
	}
		



	return IMAGE_SUCCESSFUL;
}

unsigned int treshold_MMX(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels, unsigned char lowerBound, unsigned char upperBound){
	__asm {
		MOV		EDI, pixels

		MOV		EAX, width
		IMUL	EAX, height
		IMUL	EAX, channels

		MOV		BL, lowerBound
		MOV		BH, lowerBound
		SHL		EBX, 16
		MOV		BL, lowerBound
		MOV		BH, lowerBound
		MOVD	MM1, EBX
		MOVD	MM3, EBX
		PSLLQ	MM1, 32
		POR		MM1, MM3

		MOV		BL, upperBound
		MOV		BH, upperBound
		SHL		EBX, 16
		MOV		BL, upperBound
		MOV		BH, upperBound
		MOVD	MM2, EBX
		MOVD	MM4, EBX
		PSLLQ	MM2, 32
		POR		MM2, MM4

		JMP		TheLoopFirst

		TheLoop64:
			MOVQ	MM0, [EDI]
			MOVQ	MM3, MM0
			MOVQ	MM4, MM0
			PCMPGTB	MM0, MM1
			PCMPGTB	MM3, MM2
			PCMPEQB	MM4, MM1
			POR		MM0, MM4
			PANDN	MM3, MM0
			MOVQ	[EDI], MM3

			ADD		EDI, 8
		TheLoopFirst:
			SUB		EAX, 8
		JGE	TheLoop64
		ADD	EAX, 8
		JZ	End	
		TheLoop8:
			MOV		CL, [EDI]
			CMP		CL, BL
			SETNAE	DL
			CMP		CL, BH
			SETA	DH
			OR		DL, DH
			DEC		DL			
			MOV		[EDI], DL
			INC		EDI
			DEC		EAX
		JNZ	TheLoop8

		End:

		EMMS
	}

	return IMAGE_SUCCESSFUL;
}


unsigned int duplicate(unsigned char *&dest, unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels){
	dest = (unsigned char *) malloc(width * height * channels);
	memcpy(dest, pixels, width * height * channels);
	return IMAGE_SUCCESSFUL;
}

unsigned int toGrayscale(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels){
	if (channels >= 3){
		unsigned char *dest, *src, *newPixels = (unsigned char *) malloc(width * height);
		unsigned int length = width * height;


		dest = newPixels;
		src = pixels;
		while (length--){
			*dest++ = (77 * src[0] + 151 * src[1] + 28 * src[2] + 128) >> 8;
			src += channels;
		}

		channels = 1;
		free(pixels);
		pixels = newPixels;

		return IMAGE_SUCCESSFUL;
	} else return IMAGE_IMPROPER_OPERAND;
}

int min3(int r,int g,int b){
	if (r < g){
		if (r < b) return r; else return b;
	} else {
		if (g < b) return g; else return b;
	}
}

unsigned int RGBtoHSI(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels){
	if (channels == 3){
		unsigned char *src;
		unsigned int length = width * height;
		unsigned char r,g,b,s,i;
		int rgb;
		float h;

		src = pixels;
		while (length--){
			
			r = *src;
			g = src[1];
			b = src[2];

			h = acosf( ((r-g) + (r-b)) / (2 * sqrtf(float((r-g)*(r-g) + (r-b)*(g-b)))));
			if (b > g) h = 2 * PI - h;

			rgb = r + g + b;
			s = (rgb? (255 - (1530 * min3(r,g,b) + rgb) / (2 * rgb)) : 0);
			i = rgb / 3;


			*src++ = int(255 * h / (2 * PI) + 0.5f);
			*src++ = s;
			*src++ = i;
		}

		return IMAGE_SUCCESSFUL;
	} else return IMAGE_IMPROPER_OPERAND;
}

unsigned int HSItoRGB(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels){
	if (channels == 3){
		unsigned char *src;
		unsigned int length = width * height;
		int r,g,b,h,s,i, H;
		float rt3 = 256.0f / sqrtf(3);

		src = pixels;
		while (length--){
			
			h = *src;
			s = src[1];
			i = src[2];

			if (s == 0){
				r = g = b = i;
			} else {
				if (h < 85){
					b = (255 - s) * i / 255;
					H = int(rt3 * tan(h * 2 * PI / 255 - PI / 3));
					g = ((768 + 3 * H) * i - (256 + 3 * H) * b) >> 9;
					if (g > 255) g = 255;
					r = 3 * i - g - b;
					if (r > 255) r = 255;
				} else if (h < 171){
					r = (255 - s) * i / 255;
					H = int(rt3 * tan(h * 2 * PI / 255 - PI));
					b = ((768 + 3 * H) * i - (256 + 3 * H) * r) >> 9;
					if (b > 255) b = 255;
					g = 3 * i - b - r;
					if (g > 255) g = 255;
				} else {
					g = (255 - s) * i / 255;
					H = int(rt3 * tan(h * 2 * PI / 255 - 5 * PI / 3));
					r = ((768 + 3 * H) * i - (256 + 3 * H) * g) >> 9;
					if (r > 255) r = 255;
					b = 3 * i - r - g;
					if (b > 255) b = 255;
				}
			}


			*src++ = r;
			*src++ = g;
			*src++ = b;
		}

		return IMAGE_SUCCESSFUL;
	} else return IMAGE_IMPROPER_OPERAND;
}

unsigned int addChannel(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned char *channel, unsigned int position){
	unsigned int k, length;
	unsigned char *newPixels = (unsigned char *) malloc(length = (width * height * (channels + 1)));
	unsigned char *dest, *src;
	
	dest = newPixels;
	src = pixels;
	k = 0;
	while (length--){
		*dest++ = (k == position)? *channel++ : *src++;
		k = (k == channels)? 0:k + 1;
	}
	channels++;
	
	free(pixels);
	pixels = newPixels;
	
	return IMAGE_SUCCESSFUL;
}

unsigned int addChannelWithValue(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int value, unsigned int position){
	unsigned int k, length;
	unsigned char *newPixels = (unsigned char *) malloc(length = (width * height * (channels + 1)));
	unsigned char *dest, *src;
	
	dest = newPixels;
	src = pixels;
	k = 0;
	while (length--){
		*dest++ = (k == position)? value : *src++;
		k = (k == channels)? 0:k + 1;
	}
	channels++;
	
	free(pixels);
	pixels = newPixels;
	
	return IMAGE_SUCCESSFUL;
}

unsigned int removeChannel(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int nRemove){
	if (nRemove >= 0 && nRemove < channels){
		unsigned char *newPixels = (unsigned char *) malloc(width * height * (channels - 1));
		unsigned char *dest, *src;
		unsigned int k, length = width * height * channels;
		
		channels--;
		dest = newPixels;
		src = pixels;
		k = 0;
		while (length--){
			if (k != nRemove) 
				*dest++ = *src;
			src++;
			k = (k == channels)? 0:k + 1;
		}
		
		free(pixels);
		pixels = newPixels;
		
		return IMAGE_SUCCESSFUL;
	} else return IMAGE_IMPROPER_OPERAND;
}

unsigned int flipChannels(unsigned char *pixels, unsigned int width, unsigned int height, unsigned int channels){
	if (channels == 4){
		__asm {
			
			MOV		ECX, width
			MOV		EDI, pixels
			IMUL	ECX, height
			MOV		EAX, [EDI]

			TheLoop:
				BSWAP	EAX
				MOV		[EDI], EAX
				ADD		EDI, 4
				DEC		ECX
				MOV		EAX, [EDI]
			JNZ	TheLoop
		}
	} else {
		unsigned int length = width * height;
		unsigned int i, len = channels / 2;
		unsigned char tmp, *dest = pixels;
		
		
		while (length--){
			for (i = 0; i < len; i++){
				tmp = dest[i];
				dest[i] = dest[channels - i - 1];
				dest[channels - i - 1] = tmp;
			}
			dest += channels;
		}
	}

	return IMAGE_SUCCESSFUL;
}

#define IMAGE_NEAREST  0
#define IMAGE_BILINEAR 1
#define IMAGE_BICUBIC  2

int icerp(int a, int b, int c, int d, int x){
	int p = (d - c) - (a - b);
	int q = (a - b) - p;
	int r = c - a;
	return (x * (x * (x * p + (q << 7)) + (r << 14)) + (b << 21)) >> 21;
}

void rescale(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int newWidth, unsigned int newHeight, unsigned int method = IMAGE_BILINEAR){
	unsigned char *newPixels = (unsigned char *) malloc(newWidth * newHeight * channels);
	unsigned int x,y,k,sampleX, sampleY, wX, wY;
	unsigned char *src, *dest = newPixels;

	switch(method){
	case IMAGE_NEAREST:
		for (y = 0; y < newHeight; y++){
			sampleY = (height - 1) * y / (newHeight - 1);
			for (x = 0; x < newWidth; x++){
				sampleX = (width - 1) * x / (newWidth - 1);
				for (k = 0; k < channels; k++){
					*dest++ = pixels[(sampleY * width + sampleX) * channels + k];
				}
			}
		}
		break;
	case IMAGE_BILINEAR:
		for (y = 0; y < newHeight; y++){
			sampleY = (((height - 1) * y) << 8) / (newHeight - 1);
			if (y == newHeight - 1) sampleY--;
			wY = sampleY & 0xFF;
			sampleY >>= 8;

			for (x = 0; x < newWidth; x++){
				sampleX = (((width - 1) * x) << 8) / (newWidth - 1);
				wX = sampleX & 0xFF;
				sampleX >>= 8;

				src = pixels + (sampleY * width + sampleX) * channels;

				for (k = 0; k < channels; k++){
					*dest++ = (
						(256 - wX) * (256 - wY) * src[0] + 
						(256 - wX) * (      wY) * src[width * channels] + 
						(      wX) * (256 - wY) * src[channels] + 
						(      wX) * (      wY) * src[(width + 1) * channels]) >> 16;
					*src++;
				}
			}
		}
		break;
	case IMAGE_BICUBIC:
		int a,b,c,d;
		int res;

		for (y = 0; y < newHeight; y++){
			sampleY = (((height - 1) * y) << 7) / (newHeight - 1);
			wY = sampleY & 0x7F;
			sampleY >>= 7;
			if (sampleY == 0) a = 0;
			if (sampleY >= newHeight - 1) d = 0;

			for (x = 0; x < newWidth; x++){
				sampleX = (((width - 1) * x) << 7) / (newWidth - 1);
				wX = sampleX & 0x7F;
				sampleX >>= 7;

				src = pixels + ((sampleY - 1) * width + (sampleX - 1)) * channels;

				for (k = 0; k < channels; k++){
					if (sampleY > 0){
						a = icerp(src[0], src[channels], src[2 * channels], src[3 * channels], wX);
					}
					b = icerp(src[    width * channels], src[(    width + 1) * channels], src[(    width + 2) * channels], src[(    width + 3) * channels], wX);
					c = icerp(src[2 * width * channels], src[(2 * width + 1) * channels], src[(2 * width + 2) * channels], src[(2 * width + 3) * channels], wX);
					if (sampleY < newHeight - 1){
						d = icerp(src[3 * width * channels], src[(3 * width + 1) * channels], src[(3 * width + 2) * channels], src[(3 * width + 3) * channels], wX);
					}

					res = icerp(a,b,c,d, wY);
					*dest++ = (res < 0)? 0 : (res > 255)? 255 : res;
					*src++;
				}
			}
		}
		break;
	}

	free(pixels);
	pixels = newPixels;
	width  = newWidth;
	height = newHeight;
}

void normalize(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels){
	unsigned char *src;
	unsigned int k, len;
	unsigned char min,max;

	min = max = *pixels;
	for (k = 0; k < channels; k++){
		len = width * height;

		src = pixels + k;

		do {
			src += channels;
			if (*src > max) max = *src;
			if (*src < min) min = *src;
		} while (--len);
	}

	if (max != min && (max - min < 255)){
		len = width * height * channels;
		
		src = pixels;
		do {
			*src++ = (255 * (*src - min)) / (max - min);
		} while (--len);
	}
}

void saturate(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int factor){

	RGBtoHSI(pixels,width,height,channels);

	unsigned char *src;
	unsigned int len, val;

	len = width * height;
		
	src = pixels + 1;
	do {
		val = (factor * *src);
		*src = (val > 255)? 255:val;
		src += channels;
	} while (--len);


	HSItoRGB(pixels,width,height,channels);

}

#define IMAGE_FIT      0
#define IMAGE_NORESIZE 1

void rotate(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, float angle, unsigned int mode = IMAGE_NORESIZE){
	unsigned int newWidth, newHeight;
	int Cos = int(1024 * cosf(angle));
	int Sin = int(1024 * sinf(angle));

	switch(mode){
	case IMAGE_FIT:
		newWidth  = (abs(width * Cos + height * Sin) + 512) >> 10;
		newHeight = (abs(width * Sin + height * Cos) + 512) >> 10;
		break;
	case IMAGE_NORESIZE:
		newWidth  = width;
		newHeight = height;
		break;
	}


	unsigned char *newPixels = (unsigned char *) malloc(newWidth * newHeight * channels);
	unsigned char *src, *dest = newPixels;
	unsigned int k,wX, wY;
	int x,y,minx, maxx, miny, maxy, sampleX, sampleY;

	minx = -int(newWidth >> 1);
	maxx =  int(newWidth + minx);
	miny = -int(newHeight >> 1);
	maxy =  int(newHeight + miny);


	for (y = miny; y < maxy; y++){
		for (x = minx; x < maxx; x++){

			sampleX = x * Cos - y * Sin + (width  << 9);
			sampleY = x * Sin + y * Cos + (height << 9);

			wX = sampleX & 0x3FF;
			sampleX >>= 10;
			wY = sampleY & 0x3FF;
			sampleY >>= 10;
			
			if (sampleX >= 0 && sampleX < int(width - 1) && sampleY >= 0 && sampleY < int(height - 1)){
				src = pixels + (sampleY * width + sampleX) * channels;
				
				for (k = 0; k < channels; k++){
					*dest++ = (
						(1024 - wX) * (1024 - wY) * src[0] + 
						(1024 - wX) * (       wY) * src[width * channels] + 
						(       wX) * (1024 - wY) * src[channels] + 
						(       wX) * (       wY) * src[(width + 1) * channels]) >> 20;
					*src++;
				}
			} else {
				for (k = 0; k < channels; k++){
					*dest++ = 0;
				}
			}
		}
	}

	free(pixels);
	pixels = newPixels;

	width  = newWidth;
	height = newHeight;
}

void flipX(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels){
	unsigned char *newPixels = (unsigned char *) malloc(width * height * channels);
	unsigned char *dest, *src;
	unsigned int y,k,len;

	dest = newPixels;
	for (y = 0; y < height; y++){
		src = pixels + (y + 1) * width * channels - channels;
		len = width * channels;

		k = 0;
		do {
			*dest++ = *src++;
			if (++k == channels){
				k = 0;
				src -= channels << 1;				
			}
		} while (--len);
	}
	
	free(pixels);
	pixels = newPixels;

}

void flipY(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels){
	unsigned char *newPixels = (unsigned char *) malloc(width * height * channels);
	unsigned char *dest, *src;
	unsigned int y = height;

	dest = newPixels;
	src = pixels + width * height * channels;
	do {
		src -= width * channels;
		memcpy(dest, src, width * channels);
		dest += width * channels;
	} while (--y);
	
	free(pixels);
	pixels = newPixels;

}

void transpose(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels){
	unsigned char *newPixels = (unsigned char *) malloc(width * height * channels);
	unsigned char *src, *dest = newPixels;
	unsigned int x,y,k,tmp;

	for (y = 0; y < width; y++){
		src = pixels + y * channels;
		for (x = 0; x < height; x++){
			for (k = 0; k < channels; k++){
				*dest++ = src[k];
			}
			src += width * channels;
		}
	}
	
	free(pixels);
	pixels = newPixels;

	tmp = height;
	height = width;
	width = tmp;
}

void buildMipMap(unsigned char *dest, unsigned char *src, unsigned int width, unsigned int height, unsigned int channels){
	unsigned int x,y,i,r,base0, base1;
	unsigned int xOff = (width  < 2)? 0:-1;
	unsigned int yOff = (height < 2)? 0:1;

	for (y = 0; y < height; y += 2){
		for (x = 0; x < width; x += 2){
			base0 = (y * width + x) * channels;
			base1 = ((y + yOff) * width + x) * channels;

			for (i = 0; i < channels; i++){
				r = (unsigned int(src[base0 + i]) + 
					 unsigned int(src[base0 + (xOff & channels) + i]) +
					 unsigned int(src[base1 + i]) +
					 unsigned int(src[base1 + (xOff & channels) + i]) + 2) >> 2;

				*dest = r;
				dest++;
			}
			
		}
	}
}

void mipmap(unsigned char *&pixels, unsigned int &width, unsigned int &height, unsigned int &channels, unsigned int &format){
	int w = width, h = height;
	int size = 1;

	while (w > 1 || h > 1){
		size += w * h;
		if (w > 1) w >>= 1;
		if (h > 1) h >>= 1;
		
	}

	unsigned char *src, *dest;
	dest = pixels = (unsigned char *) realloc(pixels, size * channels);

	w = width;
	h = height;

	while (w > 1 || h > 1){
		src = dest;
		dest += w * h * channels;
		buildMipMap(dest, src, w, h, channels);

		if (w > 1) w >>= 1;
		if (h > 1) h >>= 1;

	}

	format |= IMAGE_FORMAT_MIPMAPPED;
}


#endif // _IMAGEUTILS_H_