                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _TEXTURE3D_H_
#define _TEXTURE3D_H_

/* ################################################################################################################### */

#include "../3D/Vertex.h"

class Texture3D : public Texture {
protected:
	unsigned char *pixels;
public:
	Texture3D();
	virtual ~Texture3D();

	bool createVectorField(unsigned int size);
    bool createAttenuationMap(unsigned int size);

	bool upload();
};

/* ----------------------------------------------------*/

Texture3D::Texture3D(){
	pixels = NULL;
	width = height = depth = channels = 0;
	textureID = TEXTURE_INVALID;
	minFilter = GL_LINEAR;//_MIPMAP_LINEAR;
	magFilter = GL_LINEAR;
	wrapModeS = GL_CLAMP_TO_EDGE_EXT;
	wrapModeT = GL_CLAMP_TO_EDGE_EXT;
    wrapModeR = GL_CLAMP_TO_EDGE_EXT;
	internalFormat = GL_RGB8;
	target = GL_TEXTURE_3D_EXT;
}

Texture3D::~Texture3D(){
	if (pixels != NULL) delete pixels;
}

bool Texture3D::createVectorField(unsigned int size){
	if (pixels != NULL) delete pixels;

    width = height = depth = size;
    channels = 3;
    pixels = new unsigned char[size * size * size * 3];

    unsigned char *dest;
    unsigned int i,j,k,norm;
    float r, f, diff, mult;
    diff = size / 2 - 0.5f;
    mult = 1.0f / size;

    dest = pixels;
    for (k = 0; k < size; k++){
        for (j = 0; j < size; j++){
            for (i = 0; i < size; i++){
                r = mult * sqrtf(sqrf(i-diff) + sqrf(j-diff) + sqrf(k-diff));
                //f = 4 / (1 + 0.7f * r + 0.3f * r*r + 0.1f * r * r * r);
                f = (1 - 2.08f * r);

                if (f > 0.9f) f = 0.9f;
                if (f < 0) f = 0;

                Vertex n(i-diff,j-diff,k-diff);
                n.normalize();
                n = n * f;

                norm = getRGBANormal(n);
                *dest++ = norm;
                *dest++ = norm >> 8;
                *dest++ = norm >> 16;
            }
        }
    }

	return true;
}

bool Texture3D::createAttenuationMap(unsigned int size){
	if (pixels != NULL) delete pixels;

    width = height = depth = size;
    channels = 1;
    pixels = new unsigned char[size * size * size];

    unsigned char *dest;
    unsigned int i,j,k;
    float r, f, diff, mult;
    diff = size / 2 - 0.5f;
    mult = 1.0f / size;

    dest = pixels;
    for (k = 0; k < size; k++){
        for (j = 0; j < size; j++){
            for (i = 0; i < size; i++){
                r = mult * sqrtf(sqrf(i-diff) + sqrf(j-diff) + sqrf(k-diff));
                //f = 50000 / (1 + 0.7f * r + 0.3f * r*r + 0.1f * r * r * r);
                f = 255 - 530 * r;//100 / (1 + 10.0f * r + 0.3f * r * r);

                //if (r > 0.95f * diff) f = 0;
                if (f < 0) f = 0;
                //if (f > 255) f = 255;

                *dest++ = (unsigned char) f;
            }
        }
    }

	internalFormat = GL_LUMINANCE;

	return true;
}

bool Texture3D::upload(){
	if (pixels == NULL) return false;

	if (textureID == TEXTURE_INVALID){
		glGenTextures(1, &textureID);
	}
    glBindTexture(GL_TEXTURE_3D_EXT, textureID);

	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, minFilter);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, magFilter);

	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, wrapModeS);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, wrapModeT);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, wrapModeR);

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 1);
	//glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

    //upload2DImage(GL_TEXTURE_2D, pixels, width, height, channels, format, internalFormat, minFilter);
    glTexImage3DEXT(GL_TEXTURE_3D_EXT, 0, internalFormat, width, height, depth, 0, (channels == 1)? GL_LUMINANCE : GL_RGB, GL_UNSIGNED_BYTE, pixels);
	
	delete pixels;
	pixels = NULL;
	
	return true;
}


#endif // _TEXTURE2D_H_